@@ -26,4 +26,9 @@ forward <- function(){
 #' @export 
 print.forward <- function(x, ...){
   cat("Using next method")
+}
+
+#' @keywords internal
+is_forward <- function(x) {
+  inherits(x, "forward")
 }
\ No newline at end of file
