@@ -1,4 +1,5 @@
 .globals <- new.env(hash = TRUE)
+.cache_tmpls <- new.env(hash = TRUE)
 
 .onLoad <- function(libname, pkgname) {
   .globals$infoLog <- new_log(info())
@@ -8,4 +9,6 @@
   .globals$pathToPattern <- NULL
   .globals$cookiePreprocessors <- list()
   .globals$renderer <- NULL
+  .globals$cache_tmpls <- FALSE
+  .globals$wsc <- list()
 }
