@@ -1,6 +1,6 @@
 test_that("Websocket", {
   expect_error(WebsocketHandler$new())
-  ws <- WebsocketHandler$new("hello", \(msg) {
+  ws <- WebsocketHandler$new("hello", function(msg) {
     return(msg)
   })
   msg <- ws$receive(list(message = "world"))
